@@ -1472,7 +1472,7 @@ static mchunkptr mremap_chunk(p, new_size) mchunkptr p; size_t new_size;
 
   cp = (char *)mremap((char *)p - offset, size + offset, new_size, 1);
 
-  if((long)cp < 0) return 0;
+  if (cp == (char *)-1) return 0;
 
   p = (mchunkptr)(cp + offset);
 
