@@ -1,5 +1,5 @@
 #if 0
-static char rcsid[] = "@(#)msgcat.c,v 1.1.1.1.2.1.2.1 1995/04/29 17:45:05 hjl Exp";
+static char rcsid[] = "@(#)msgcat.c,v 1.3 1995/04/15 07:36:13 hjl Exp";
 #endif
 
 /* -*- c++ -*- */
