@@ -12,9 +12,8 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.  */
 
 #include <ansidecl.h>
-#include "mathl.h"
 #include <errno.h>
-#include <fp.h>
+#include <math.h>
 
 static inline long double
 domain (long double x)
