@@ -23,7 +23,7 @@ long double fabsl(), floorl();
 
 /*  Returns fractional part of d, stores integer part in *integ
  */
-long double modf(long double d, long double *integ)
+long double modfl(long double d, long double *integ)
 {
   long double i, f;
 
