@@ -2,8 +2,12 @@
 #include <sys/socket.h>
 #include <sys/socketcall.h>
 
+#ifdef __SVR4_I386_ABI_L1__
+#define socketcall __socketcall
+#else
 static inline
 _syscall2(long,socketcall,int,call,unsigned long *,args);
+#endif
 
 /* shutdown by bir7@leland.stanford.edu */
 int
@@ -14,3 +18,7 @@ shutdown (int sockfd, int how)
   args[1] = how;
   return (socketcall (SYS_SHUTDOWN, args));
 }
+
+#ifdef __ELF__
+#pragma weak shutdown
+#endif
